#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from __future__ import annotations
import bisect
from dataclasses import dataclass
from typing import Optional, List, Tuple, cast
from typing import final
from functools import cache
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
from pyhope.mesh.mesh_vars import SIDE
# ----------------------------------------------------------------------------------------------------------------------------------
RED   = True
BLACK = False
LEFT  = True
RIGHT = False
# ==================================================================================================================================


@final
class LinkOffsetManager:
    """ Batch Update Manager for Connection Offsets

        Instead of updating each node's stored link on every insertion, this manager maintains breakpoints for cumulative index
        shifts
    """

    __slots__ = ('breakpoints',)

    def __init__(self) -> None:
        # Single breakpoint at stored_index=0 => offset=0
        self.breakpoints: List[Tuple[int, int]] = [(0, 0)]

    def update(self,
               effective_index: int,
               delta          : int) -> None:
        """ Record that all stored links with value >= insert_index should be increased by delta

            Perform a binary search on breakpoints, then update or insert the breakpoint and adjust all subsequent offsets
        """
        # S = E - offset(S)
        insert_index = effective_index - self.get_offset(effective_index)
        pos          = bisect.bisect_left(self.breakpoints, (insert_index, -float('inf')))

        # If there is a breakpoint exactly at insert_index, update its offset
        if pos < len(self.breakpoints) and self.breakpoints[pos][0] == insert_index:
            index, current_offset = self.breakpoints[pos]
            self.breakpoints[pos] = (index, current_offset + delta)
            pos += 1
        # Otherwise, insert a new breakpoint
        else:
            prev_offset = self.breakpoints[pos - 1][1] if pos > 0 else 0
            self.breakpoints.insert(pos, (insert_index, prev_offset + delta))
            pos += 1

        # INFO: This is the original implementation
        # > Directly update the breakpoints list
        # # Increase offset of all subsequent breakpoints
        # for i in range(pos, len(self.breakpoints)):
        #     index, current_offset = self.breakpoints[i]
        #     # Calculate the new offset by adding delta to the current offset
        #     self.breakpoints[i] = (index, current_offset + delta)

        # NOTE: This is the alternative implementation
        # > Convert the slice of breakpoints to a NumPy array
        breakpoints_array = np.array(self.breakpoints[pos:], dtype=[('index', int), ('offset', int)]).reshape(-1, 2)
        # Add delta to the 'offset' column
        breakpoints_array['offset'] += delta
        # Convert back to a list of tuples
        self.breakpoints[ pos:    ]  = cast(list[tuple[int, int]], breakpoints_array.tolist())

        # Clear any cached offset computations
        self.get_offset.cache_clear()

    @cache
    def get_offset(self,
                   index: int) -> int:
        """ Cached offset lookup for stored index.
        """
        # Binary search for the region containing 'index'
        pos = bisect.bisect_right(self.breakpoints, (index, float('inf'))) - 1
        if pos < 0:
            # If index < the first breakpoint, offset=0
            return 0
        return self.breakpoints[pos][1]


@final
@dataclass(init=False, repr=False, eq=False, slots=False)
class SideNode:
    # PERF: Convert class to dataclass for better performance
    # """
    # value: a SIDE object
    # link : the stored connection (an int) from the SIDE (side.connection)
    # """
    # value = SIDE
    # link  = Optional[int]  # This is the base (stored) connection value
    __slots__ = ('value', 'link')

    def __init__(self,
                 value: SIDE,
                 link: Optional[int]) -> None:
        """
        value: a SIDE object
        link : the stored connection (an int) from the SIDE (side.connection)
        """
        self.value = value
        self.link  = link   # This is the base (stored) connection value

    def effective_link(self,
                       offset_manager: LinkOffsetManager) -> Optional[int]:
        """
        Compute the effective connection (link) by adding the current offset
        """
        if self.link is None:
            return None
        return cast(int, self.link) + offset_manager.get_offset(self.link)


# ----- Red-Black Tree Helpers ---------------------------------------------------------------------------
@final
@dataclass(init=True, repr=False, eq=False, slots=True)
class _RBTreeNode:
    # PERF: Convert class to dataclass for better performance
    # __slots__ = ('data', 'color', 'left', 'right', 'parent', 'size')
    #
    # def __init__(self,
    #              data: SideNode) -> None:
    #     self.data  : SideNode = data
    #     self.color : bool     = RED  # New nodes are initially red
    #     self.left  : Optional[_RBTreeNode] = None
    #     self.right : Optional[_RBTreeNode] = None
    #     self.parent: Optional[_RBTreeNode] = None
    #     self.size  : int = 1
    data  : SideNode
    color : bool     = RED  # New nodes are initially red
    left  : Optional[_RBTreeNode] = None
    right : Optional[_RBTreeNode] = None
    parent: Optional[_RBTreeNode] = None
    size  : int = 1


@final
class RedBlackTree:
    """
    This class provides a balanced binary search tree implemented as a Red-Black Tree,
    augmented with subtree sizes to support efficient arbitrary insertions and random access.
    """
    __slots__ = ('_root', '_size', 'offset_manager', '_node_at')

    def __init__(self,
                 offset_manager: LinkOffsetManager) -> None:
        self._root: Optional[_RBTreeNode] = None
        self._size: int = 0
        # Shared manager for batch updates
        self.offset_manager = offset_manager
        # Cache for node_at lookups using native cache
        self._node_at = self._node_at_impl

    def __len__(self) -> int:
        return self._size

    def _left_rotate(self,
                     x: _RBTreeNode) -> None:
        """ Rotate the subtree rooted at x to the left
        """
        y = x.right

        # Cannot left-rotate without right child
        if y is None:
            return

        x.right  = y.left
        if y.left is not None:
            y.left.parent = x
        y.parent = x.parent

        if x.parent is None:
            self._root = y
        elif x == x.parent.left:
            x.parent.left = y
        else:
            x.parent.right = y

        y.left   = x
        x.parent = y
        # Update subtree sizes
        x.size   = 1 + (x.left.size if x.left else 0) + (x.right.size if x.right else 0)
        y.size   = 1 + (y.left.size if y.left else 0) + (y.right.size if y.right else 0)

    def _right_rotate(self,
                      x: _RBTreeNode) -> None:
        """ Rotate the subtree rooted at x to the right
        """
        y = x.left

        # Cannot right-rotate without left child
        if y is None:
            return

        x.left   = y.right
        if y.right is not None:
            y.right.parent = x
        y.parent = x.parent

        if x.parent is None:
            self._root = y
        elif x == x.parent.right:
            x.parent.right = y
        else:
            x.parent.left = y

        y.right  = x
        x.parent = y
        # Update subtree sizes
        x.size   = 1 + (x.left.size if x.left else 0) + (x.right.size if x.right else 0)
        y.size   = 1 + (y.left.size if y.left else 0) + (y.right.size if y.right else 0)

    def _insert_fixup(self, z: _RBTreeNode) -> None:
        """ Fixup the red-black tree after insertion
        """
        while z.parent is not None and z.parent.color == RED:
            if z.parent == z.parent.parent.left:
                y = z.parent.parent.right

                if y is not None and y.color == RED:
                    z.parent.color = BLACK
                    y.color        = BLACK
                    z.parent.parent.color = RED
                    z = cast(_RBTreeNode, z.parent.parent)
                else:
                    if z == z.parent.right:
                        z = z.parent
                        self._left_rotate(z)
                    z.parent.color = BLACK
                    z.parent.parent.color = RED
                    self._right_rotate(cast(_RBTreeNode, z.parent.parent))

            else:
                y = z.parent.parent.left
                if y is not None and y.color == RED:
                    z.parent.color = BLACK
                    y.color        = BLACK
                    z.parent.parent.color = RED
                    z = cast(_RBTreeNode, z.parent.parent)
                else:
                    if z == z.parent.left:
                        z = z.parent
                        self._right_rotate(z)
                    z.parent.color = BLACK
                    z.parent.parent.color = RED
                    self._left_rotate(cast(_RBTreeNode, z.parent.parent))
        self._root.color = BLACK

    @cache
    def _node_at_impl(self, index: int) -> SideNode:
        if not 0 <= index < self._size:
            raise IndexError('Index out of range')

        node = self._root
        while node is not None:
            left_size = node.left.size if node.left else 0
            if index < left_size:
                node = node.left
            elif index == left_size:
                return node.data
            else:
                index -= left_size + 1
                node = node.right
        raise IndexError("Index not found")

    def node_at(self, index: int) -> SideNode:
        """
        Retrieve the node at the given index (via cached lookup)
        """
        return self._node_at(index)

    def __getitem__(self, index: int) -> SideNode:
        return self.node_at(index)

    def insert(self, effective_index: int, new_node: SideNode, update_offset: bool = True) -> None:
        """
        Insert new_node at the logical position corresponding to the effective_index in the red-black tree.
        """
        if not 0 <= effective_index <= self._size:
            raise IndexError('Index out of range')

        # Invalidate the node_at cache since the tree structure is about to change
        if update_offset:
            self._node_at.cache_clear()

        # Create a new red-black tree node for new_node
        z: _RBTreeNode           = _RBTreeNode(new_node)
        y: Optional[_RBTreeNode] = None
        x: Optional[_RBTreeNode] = self._root
        current_index: int       = effective_index
        branch: Optional[bool]   = None

        # Find the insertion point using order statistics
        while x is not None:
            y         = x
            x.size   += 1  # Increment subtree size for nodes along the insertion path
            left_size = x.left.size if x.left is not None else 0

            if current_index <= left_size:
                branch = LEFT
                x = x.left
            else:
                current_index -= left_size + 1
                branch = RIGHT
                x = x.right

        z.parent = y
        if y is None:
            self._root = z
        else:
            if branch == LEFT:
                y.left  = z
            else:
                y.right = z

        # Fix-up the red-black tree properties
        self._insert_fixup(z)
        self._size += 1

        if update_offset:
            self.offset_manager.update(effective_index, 1)

    def update(self, index: int, new_value: SIDE) -> None:
        """
        Update the value of the node at the given index with the new SIDE object
        """
        node = self.node_at(index)
        node.value = new_value

    def inorder(self, t: Optional[_RBTreeNode], result: List[SideNode]) -> None:
        """
        Recursively traverse the red-black tree in order and append node data to the result list
        """
        if t is None:
            return
        self.inorder( t.left, result)
        result.append(t.data)
        self.inorder( t.right, result)

    def _to_list(self) -> List[SideNode]:
        """
        Return a Python list of the nodes (in order) via an in-order traversal of the red-black tree
        """
        result: List[SideNode] = []
        self.inorder(self._root, result)
        return result

    def __iter__(self):
        """
        Iterate over the nodes in order
        """
        return iter(self.to_list())

    @staticmethod
    def _build_tree(nodes: tuple[_RBTreeNode, ...], start: int, end: int) -> Optional[_RBTreeNode]:
        """Recursively build a balanced tree from the tuple of nodes

        This method assumes that the nodes are in the desired in-order sequence
        """
        # Return if subtree is empty
        if start > end:
            return None

        mid  = (start + end) // 2
        root = nodes[mid]

        # Recursively build left and right subtrees
        root.left  = RedBlackTree._build_tree(nodes, start, mid - 1)
        if root.left is not None:
            root.left.parent  = root

        root.right = RedBlackTree._build_tree(nodes, mid + 1, end)
        if root.right is not None:
            root.right.parent = root

        # Compute subtree size in one pass
        left_size  = root.left.size  if root.left  is not None else 0  # noqa: E272
        right_size = root.right.size if root.right is not None else 0  # noqa: E272
        root.size  = 1 + left_size + right_size

        return root

    @classmethod
    def from_list(cls, sides: List[SIDE], offset_manager: LinkOffsetManager) -> RedBlackTree:
        """Optimized bulk conversion from a sorted list to a red–black tree

        Assumes that the provided list is already in the desired in-order sequence.
        All nodes are initialized as black to maintain red–black properties.
        """
        rbt   = cls(offset_manager)

        # Create a list of nodes with the SIDE data
        nodes = tuple(_RBTreeNode(SideNode(value=side, link=side.connection), color=BLACK) for side in sides)

        rbt._root = cls._build_tree(nodes, 0, len(nodes) - 1)
        rbt._size = len(nodes)

        return rbt

    # PERF: This is an improved implementation but still not optimal
    # def from_list(cls, sides: List[SIDE], offset_manager: LinkOffsetManager) -> RedBlackTree:
    #     """ Convert a list of SIDE objects into a red-black tree (balanced BST) for efficient insertion and random access
    #
    #         When building from an existing list, we do not update offsets because the offset manager because the stored
    #         connection (side.connection) is already valid
    #     """
    #     rbt = cls(offset_manager)
    #     for side in sides:
    #         node = SideNode(value=side, link=side.connection)
    #         # Do not update the offset manager during this bulk conversion
    #         rbt.insert(len(rbt), node, update_offset=False)
    #     return rbt

    # INFO: This is the original implementation
    # def to_list(self) -> List[SIDE]:
    #     """ Convert the red-black tree (balanced BST) back into a list of SIDE objects
    #
    #         For each node, update its SIDE object's connection field using the effective link, and update sideID to reflect
    #         the node's new position in the red-black tree.
    #     """
    #     nodes = self._to_list()
    #     for idx, node in enumerate(nodes):
    #         if node.value.connection is not None and node.value.connection >= 0:
    #             node.value.connection = node.effective_link(self.offset_manager) if node.link is not None else None
    #         node.value.sideID = idx
    #     return [node.value for node in nodes]

    def to_list(self) -> List[SIDE]:
        """ Convert the red-black tree back into a list of SIDE objects

        Instead of computing each node's effective link individually (using a binary search for each call), we batch-process updates
        by precomputing the breakpoints and then sweeping through the nodes that require an update.
        """
        nodes = self._to_list()

        # Extract the current breakpoints from the offset manager.
        # Assumed to be sorted by the stored index.
        breakpoints = self.offset_manager.breakpoints

        # Create a list of tuples containing (stored_link, SideNode)
        update_nodes: list[tuple[int, SideNode]] = []

        # Only consider nodes with a valid stored connection (>= 0)
        for node in nodes:
            # We use the stored link (node.link) for the update computation
            if node.link is not None and node.link >= 0:
                update_nodes.append((node.link, node))

        # Sort the update_nodes by the stored connection value
        update_nodes.sort(key=lambda tup: tup[0])

        bp_index = 0
        bp_count = len(breakpoints)
        # Iterate over the nodes (sorted by connection) and assign the effective link
        for stored_link, node in update_nodes:
            # Advance the breakpoint pointer while the next breakpoint index is <= stored_link
            while bp_index < bp_count - 1 and breakpoints[bp_index + 1][0] <= stored_link:
                bp_index += 1
            offset = breakpoints[bp_index][1]
            node.value.connection = stored_link + offset

        # Update the sideID for all nodes based on in-order position
        for idx, node in enumerate(nodes):
            node.value.sideID = idx

        return [node.value for node in nodes]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (C) 2022 Nico Schlömer
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def gmsh_to_meshio(gmsh) -> meshio.Mesh:
    """
    Convert a Gmsh object to a meshio object.
    """
    # Local imports ----------------------------------------
    from pyhope.meshio.meshio_ordering import NodeOrdering
    # ------------------------------------------------------

    # Initialize the node ordering
    node_ordering = NodeOrdering()

    # Extract point coords
    idx, points, _ = gmsh.model.mesh.getNodes()
    points  = np.asarray(points).reshape(-1, 3)
    idx    -= 1
    srt     = np.argsort(idx)
    assert np.all(idx[srt] == np.arange(len(idx)))
    points = points[srt]

    # Extract cells
    elem_types, elem_tags, node_tags = gmsh.model.mesh.getElements()
    cells = []
    for elem_type, elem_tags, node_tags in zip(elem_types, elem_tags, node_tags):
        # `elementName', `dim', `order', `numNodes', `localNodeCoord', `numPrimaryNodes'
        num_nodes_per_cell = gmsh.model.mesh.getElementProperties(elem_type)[3]

        node_tags_reshaped = np.asarray(node_tags).reshape(-1, num_nodes_per_cell) - 1
        node_tags_reshaped = node_ordering.ordering_gmsh_to_meshio(elem_type, node_tags_reshaped)

        # NRG: Fix the element ordering
        node_tags_sorted   = node_tags_reshaped[np.argsort(elem_tags)]
        cells.append(meshio.CellBlock(meshio.gmsh.gmsh_to_meshio_type[elem_type], node_tags_sorted))

    cell_sets = {}
    for dim, tag in gmsh.model.getPhysicalGroups():
        # Get offset of the node tags (gmsh sorts elements of all dims in succeeding order of node tags, but order of dims might differ)
        _, elem_tags, _ = gmsh.model.mesh.getElements(dim=dim)
        offset = min(elem_tags[0])
        
        name = gmsh.model.getPhysicalName(dim, tag)
        cell_sets[name] = [[] for _ in range(len(cells))]
        for e in gmsh.model.getEntitiesForPhysicalGroup(dim, tag):
            # elem_types, elem_tags, node_tags
            elem_types, elem_tags, _ = gmsh.model.mesh.getElements(dim, e)
            assert len(elem_types) == len(elem_tags)
            assert len(elem_types) == 1
            elem_type = elem_types[0]
            elem_tags = elem_tags[0] - offset

            meshio_cell_type = meshio.gmsh.gmsh_to_meshio_type[elem_type]
            # Make sure that the cell type appears only once in the cell list
            idx = []
            for k, cell_block in enumerate(cells):
                if cell_block.type == meshio_cell_type:
                    idx.append(k)
            assert len(idx) == 1
            idx = idx[0]
            cell_sets[name][idx].append(elem_tags)

        cell_sets[name] = [(None if len(idcs) == 0 else np.concatenate(idcs)) for idcs in cell_sets[name]]

    return meshio.Mesh(points, cells, cell_sets=cell_sets)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
from functools import cache
from typing import Union, Tuple, Any
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
import numpy.typing as npt
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.mesh.mesh_vars as mesh_vars
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# Instantiate ELEMTYPE
elemTypeClass = mesh_vars.ELEMTYPE()
# ==================================================================================================================================


@cache
def faces(elemType: Union[int, str]) -> list[str]:
    """ Return a list of all sides of an element
    """
    faces_map = {  # Tetrahedron
                   4: ['z-', 'y-', 'x+', 'x-'            ],
                   # Pyramid
                   5: ['z-', 'y-', 'x+', 'y+', 'x-'      ],
                   # Wedge / Prism
                   6: ['y-', 'x+', 'x-', 'z-', 'z+'      ],
                   # Hexahedron
                   8: ['z-', 'y-', 'x+', 'y+', 'x-', 'z+']
                }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in faces: elemType {elemType} is not supported')

    return faces_map[elemType % 100]


@cache
def edges(elemType: Union[int, str]) -> list[int]:
    """ Return a list of all edges of an element
    """
    edges_map = {  # Tetrahedron
                   4: [0, 1, 2, 3, 4, 5],
                   # Pyramid
                   5: [0, 1, 2, 3, 4, 5, 6, 7],
                   # Wedge / Prism
                   6: [0, 1, 2, 3, 4, 5, 6, 7, 8],
                   # Hexahedron
                   8: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
                }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in edges_map:
        raise ValueError(f'Error in edges: elemType {elemType} is not supported')

    return edges_map[elemType % 100]


@cache
def edge_to_dir(edge: int, elemType: Union[int, str]) -> int:
    """ GMSH: Create edges from points in the given direction
    """
    eps = np.finfo(np.float64).eps
    dir_map  = {  # Tetrahedron
                  # Pyramid
                  # Wedge / Prism
                  # Hexahedron
                  8: {  0:  eps,  2:  eps,  4: eps,  6:   eps,  # Direction 0
                        1:   1.,  3:   1.,  5:  1.,  7:    1.,  # Direction 1
                        8:   2.,  9:   2., 10:  2., 11:    2.}  # Direction 2
               }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in dir_map:
        raise ValueError(f'Error in edge_to_direction: elemType {elemType} is not supported')

    dir = dir_map[elemType % 100]

    try:
        return (np.rint(abs(dir[edge]))).astype(int)
    except KeyError:
        raise KeyError(f'Error in edge_to_dir: edge {edge} is not supported')


@cache
def edge_to_corner(edge: int, elemType: Union[int, str], dtype=int) -> np.ndarray:
    """ GMSH: Get points on edges
    """
    edge_map = {  # Tetrahedron
                  4: [ [0, 1], [1, 2], [2, 1], [0, 3],
                       [1, 3], [2, 3]                 ],
                  # Pyramid
                  5: [ [0, 1], [1, 2], [2, 3], [3, 0],
                       [0, 4], [1, 5], [2, 4], [3, 4] ],
                  # Wedge / Prism
                  6: [ [0, 1], [1, 2], [2, 0], [0, 3],
                       [2, 3], [3, 4], [4, 5], [5, 4] ],
                  # Hexahedron
                  8: [ [0, 1], [1, 2], [2, 3], [3, 0],
                       [0, 4], [1, 5], [2, 6], [3, 7],
                       [4, 5], [5, 6], [6, 7], [7, 4] ],
               }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in edge_map:
        raise ValueError(f'Error in edge_to_corner: elemType {elemType} is not supported')

    edges = edge_map[elemType % 100]

    try:
        return np.array(edges[edge], dtype=dtype)
    except KeyError:
        raise KeyError(f'Error in edge_to_corner: edge {edge} is not supported')


@cache
def face_to_edge(face: str, elemType: Union[str, int], dtype=int) -> np.ndarray:
    """ GMSH: Create faces from edges in the given direction
    """
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: {  'z-': np.array((  0,  1,   2,   3), dtype=dtype),
                         'y-': np.array((  0,  9,  -4,  -8), dtype=dtype),
                         'x+': np.array((  1, 10,  -5,  -9), dtype=dtype),
                         'y+': np.array(( -2, 10,   6, -11), dtype=dtype),
                         'x-': np.array((  8, -7, -11,   3), dtype=dtype),
                         'z+': np.array((  4,  5,   6,   7), dtype=dtype)}
                }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_edge: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][face]
    except KeyError:
        raise KeyError(f'Error in face_to_edge: face {face} is not supported')


@cache
def face_to_corner(face, elemType: Union[str, int], dtype=int) -> np.ndarray:
    """ GMSH: Get points on faces in the given direction
    """
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: {  'z-': np.array((  0,  1,   2,   3), dtype=dtype),
                         'y-': np.array((  0,  1,   5,   4), dtype=dtype),
                         'x+': np.array((  1,  2,   6,   5), dtype=dtype),
                         'y+': np.array((  2,  6,   7,   3), dtype=dtype),
                         'x-': np.array((  0,  4,   7,   3), dtype=dtype),
                         'z+': np.array((  4,  5,   6,   7), dtype=dtype)}
                }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_corner: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][face]
    except KeyError:
        raise KeyError(f'Error in face_to_corner: face {face} is not supported')


@cache
def face_to_cgns(face: str, elemType: Union[str, int], dtype=int) -> np.ndarray:
    """ CGNS: Get points on faces in the given direction
    """
    faces_map = {  # Tetrahedron
                   4: {'z-': np.array((  0,  2,  1    ), dtype=dtype),
                       'y-': np.array((  0,  1,  3    ), dtype=dtype),
                       'x+': np.array((  1,  2,  3    ), dtype=dtype),
                       'x-': np.array((  2,  0,  3    ), dtype=dtype)},
                   # Pyramid
                   5: {'z-': np.array((  0,  3,  2,  1), dtype=dtype),
                       'y-': np.array((  0,  1,  4    ), dtype=dtype),
                       'x+': np.array((  1,  2,  4    ), dtype=dtype),
                       'y+': np.array((  2,  3,  4    ), dtype=dtype),
                       'x-': np.array((  3,  0,  4    ), dtype=dtype)},
                   # Wedge / Prism
                   6: {'y-': np.array((  0,  1,  4,  3), dtype=dtype),
                       'x+': np.array((  1,  2,  5,  4), dtype=dtype),
                       'x-': np.array((  2,  0,  3,  5), dtype=dtype),
                       'z-': np.array((  0,  2,  1    ), dtype=dtype),
                       'z+': np.array((  3,  4,  5    ), dtype=dtype)},
                   # Hexahedron
                   8: {'z-': np.array((  0,  3,  2,  1), dtype=dtype),
                       'y-': np.array((  0,  1,  5,  4), dtype=dtype),
                       'x+': np.array((  1,  2,  6,  5), dtype=dtype),
                       'y+': np.array((  2,  3,  7,  6), dtype=dtype),
                       'x-': np.array((  0,  4,  7,  3), dtype=dtype),
                       'z+': np.array((  4,  5,  6,  7), dtype=dtype)}
                }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_cgns: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][face]
    except KeyError:
        raise KeyError(f'Error in face_to_cgns: face {face} is not supported')


@cache
def flip_s2m(N: int, flip: int) -> np.ndarray:
    # Create grid index arrays for the rows and columns
    p = np.arange(N)
    q = np.arange(N)

    # Create a meshgrid of row (p) and column (q) indices
    p_grid, q_grid = np.meshgrid(p, q)

    # Map row and column indices based on flip logic
    # WARNING: FOR SOME REASON, ONLY FLIP 1,3,4 IS USED WITH FACE_TO_NODES
    if flip == 0:
        return np.stack((q_grid        , p_grid        ), axis=-1)
    elif flip == 1:
        return np.stack((p_grid        , q_grid        ), axis=-1)
    elif flip == 2:
        return np.stack((N - p_grid - 1, q_grid        ), axis=-1)
    elif flip == 3:
        return np.stack((N - p_grid - 1, N - q_grid - 1), axis=-1)
    elif flip == 4:
        return np.stack((N - q_grid - 1, p_grid        ), axis=-1)
    else:
        raise ValueError('Flip must be an integer between 0 and 4')


@cache
def type_to_mortar_flip(elemType: Union[int, str]) -> dict[int, dict[int, int]]:
    """ Returns the flip map for a given element type
    """

    flipID_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: { 0: {1: 1, 2: 2, 3: 3, 4: 4},
                        1: {1: 2, 4: 1, 3: 4, 2: 3},
                        2: {3: 1, 4: 2, 1: 3, 2: 4},
                        3: {2: 1, 3: 2, 4: 3, 1: 4}}
                }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in flipID_map:
        raise ValueError(f'Error in type_to_mortar_flip: elemType {elemType} is not supported')

    try:
        return flipID_map[elemType % 100]
    except KeyError:
        raise KeyError(f'Error in type_to_mortar_flip: elemType {elemType} is not supported')


@cache
def face_to_nodes(face: str, elemType: int, nGeo: int) -> np.ndarray:
    """ Returns the tensor-product nodes associated with a face
    """
    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    order     = nGeo
    # https://hopr.readthedocs.io/en/latest/_images/CGNS_edges.jpg
    # faces_map = {  # Tetrahedron
    #                4: {  # Sides aligned with the axes
    #                      'z-': [s for s  in LINMAP(104 if order == 1 else 204, order=order)[:    , :    , 0    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      'y-': [s for s  in LINMAP(104 if order == 1 else 204, order=order)[:    , 0    , :    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      'x-': [s for s  in LINMAP(104 if order == 1 else 204, order=order)[0    , :    , :    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      # Diagonal side
    #                      'x+': [s for i in range(order+1)
    #                               for j in range(order+1-i) for s in [LINMAP(104 if order == 1 else 204, order=order)[i, j, order-i-j]] if s != -1]},  # noqa: E272, E501
    #                # Pyramid
    #                5: {  # Sides aligned with the axes
    #                      'z-': [s for s  in LINMAP(105 if order == 1 else 205, order=order)[:    , :    , 0    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      'y-': [s for s  in LINMAP(105 if order == 1 else 205, order=order)[:    , 0    , :    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      'x-': [s for s  in LINMAP(105 if order == 1 else 205, order=order)[0    , :    , :    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      # Diagonal sides
    #                      'x+': [s for i  in range(order+1) for s in LINMAP(105 if order == 1 else 205, order=order)[:, order-i, i] if s != -1],   # noqa: E272, E501
    #                      'y+': [s for i  in range(order+1) for s in LINMAP(105 if order == 1 else 205, order=order)[order-i, :, i] if s != -1]},  # noqa: E272, E501
    #                # Wedge
    #                6: {  # Sides aligned with the axes
    #                      'y-': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[:    , 0    , :    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      'x-': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[0    , :    , :    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      'z-': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[:    , :    , 0    ].flatten()         if s != -1],   # noqa: E272, E501
    #                      'z+': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[:    , :    , order].flatten()         if s != -1],   # noqa: E272, E501
    #                      # Diagonal side
    #                      'x+': [s for i  in range(order+1) for s in LINMAP(106 if order == 1 else 206, order=order)[i, order-i, :] if s != -1]},  # noqa: E272, E501
    #                # Hexahedron
    #                8: {  'z-':              LINMAP(108 if order == 1 else 208, order=order)[:    , :    , 0    ] ,                                # noqa: E272, E501
    #                      'y-': np.transpose(LINMAP(108 if order == 1 else 208, order=order)[:    , 0    , :    ]),                                # noqa: E272, E501
    #                      'x+': np.transpose(LINMAP(108 if order == 1 else 208, order=order)[order, :    , :    ]),                                # noqa: E272, E501
    #                      'y+':              LINMAP(108 if order == 1 else 208, order=order)[:    , order, :    ] ,                                # noqa: E272, E501
    #                      'x-':              LINMAP(108 if order == 1 else 208, order=order)[0    , :    , :    ] ,                                # noqa: E272, E501
    #                      'z+': np.transpose(LINMAP(108 if order == 1 else 208, order=order)[:    , :    , order])}                                # noqa: E272, E501
    #
    #             }
    # if elemType % 100 not in faces_map:
    #     raise ValueError(f'Error in face_to_nodes: elemType {elemType} is not supported')
    #
    # try:
    #     return faces_map[elemType % 100][face]
    # except KeyError:
    #     raise KeyError(f'Error in face_to_cgns: face {face} is not supported')

    match elemType % 100:
        case 4:  # Tetrahedron
            faces_map = {  # Sides aligned with the axes
                           'z-': [s for s  in LINMAP(104 if order == 1 else 204, order=order)[:    , :    , 0    ].flatten()         if s != -1],   # noqa: E272, E501
                           'y-': [s for s  in LINMAP(104 if order == 1 else 204, order=order)[:    , 0    , :    ].flatten()         if s != -1],   # noqa: E272, E501
                           'x-': [s for s  in LINMAP(104 if order == 1 else 204, order=order)[0    , :    , :    ].flatten()         if s != -1],   # noqa: E272, E501
                           # Diagonal side
                           'x+': [s for i in range(order+1)
                                    for j in range(order+1-i) for s in [LINMAP(104 if order == 1 else 204, order=order)[i, j, order-i-j]] if s != -1]    # noqa: E272, E501
                        }
        case 5:  # Pyramid
            faces_map = {  # Sides aligned with the axes
                           'z-': [s for s  in LINMAP(105 if order == 1 else 205, order=order)[:    , :    , 0    ].flatten()         if s != -1],   # noqa: E272, E501
                           'y-': [s for s  in LINMAP(105 if order == 1 else 205, order=order)[:    , 0    , :    ].flatten()         if s != -1],   # noqa: E272, E501
                           'x-': [s for s  in LINMAP(105 if order == 1 else 205, order=order)[0    , :    , :    ].flatten()         if s != -1],   # noqa: E272, E501
                           # Diagonal sides
                           'y+': [s for i  in range(order+1) for s in LINMAP(105 if order == 1 else 205, order=order)[:, order-i, i] if s != -1],   # noqa: E272, E501
                           'x+': [s for i  in range(order+1) for s in LINMAP(105 if order == 1 else 205, order=order)[order-i, :, i] if s != -1]    # noqa: E272, E501
                        }
        case 6:  # Wedge
            faces_map = {  # Sides aligned with the axes
                           'y-': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[:    , 0    , :    ].flatten()         if s != -1],   # noqa: E272, E501
                           'x-': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[0    , :    , :    ].flatten()         if s != -1],   # noqa: E272, E501
                           'z-': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[:    , :    , 0    ].flatten()         if s != -1],   # noqa: E272, E501
                           'z+': [s for s  in LINMAP(106 if order == 1 else 206, order=order)[:    , :    , order].flatten()         if s != -1],   # noqa: E272, E501
                           # Diagonal side
                           'x+': [s for i  in range(order+1) for s in LINMAP(106 if order == 1 else 206, order=order)[i, order-i, :] if s != -1]    # noqa: E272, E501
                        }
        case 8:  # Hexahedron
            faces_map = {  'z-':              LINMAP(108 if order == 1 else 208, order=order)[:    , :    , 0    ] ,                                # noqa: E272, E501
                           'y-': np.transpose(LINMAP(108 if order == 1 else 208, order=order)[:    , 0    , :    ]),                                # noqa: E272, E501
                           'x+': np.transpose(LINMAP(108 if order == 1 else 208, order=order)[order, :    , :    ]),                                # noqa: E272, E501
                           'y+':              LINMAP(108 if order == 1 else 208, order=order)[:    , order, :    ] ,                                # noqa: E272, E501
                           'x-':              LINMAP(108 if order == 1 else 208, order=order)[0    , :    , :    ] ,                                # noqa: E272, E501
                           'z+': np.transpose(LINMAP(108 if order == 1 else 208, order=order)[:    , :    , order])                                 # noqa: E272, E501
                        }
        case _:
            raise ValueError(f'Error in face_to_nodes: elemType {elemType} is not supported')

    try:
        return faces_map[face]
    except KeyError:
        raise KeyError(f'Error in face_to_nodes: face {face} is not supported')


@cache
def dir_to_nodes(dir: str, elemType: Union[str, int], nGeo: int) -> Tuple[Any, bool]:
    """ Returns the tensor-product nodes associated with a face
    """
    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    # FIXME: check for non-hexahedral elements
    order     = nGeo
    faces_map = {  # Tetrahedron
                   4: { 'z-': ((slice(None), slice(None), 0          ), False),   #              elemNodes[:    , :    , 0    ],  # noqa: E262, E501
                        'y-': ((slice(None), 0          , slice(None)), True ),   # np.transpose(elemNodes[:    , 0    , :    ]), # noqa: E262, E501
                        'x+': ((order      , slice(None), slice(None)), True ),   # np.transpose(elemNodes[order, :    , :    ]), # noqa: E262, E501
                        'x-': ((0          , slice(None), slice(None)), False)},  #              elemNodes[0    , :    , :    ]}, # noqa: E262, E501
                   # Pyramid
                   5: { 'z-': ((slice(None), slice(None), 0          ), False),   #              elemNodes[:    , :    , 0    ],  # noqa: E262, E501
                        'y-': ((slice(None), 0          , slice(None)), True ),   # np.transpose(elemNodes[:    , 0    , :    ]), # noqa: E262, E501
                        'x+': ((order      , slice(None), slice(None)), True ),   # np.transpose(elemNodes[order, :    , :    ]), # noqa: E262, E501
                        'y+': ((slice(None), order      , slice(None)), False),   #              elemNodes[:    , order, :    ],  # noqa: E262, E501
                        'x-': ((0          , slice(None), slice(None)), False)},  #              elemNodes[0    , :    , :    ]}, # noqa: E262, E501
                   # Wedge / Prism
                   6: { 'z-': ((slice(None), slice(None), 0          ), False),   #              elemNodes[:    , :    , 0    ],  # noqa: E262, E501
                        'y-': ((slice(None), 0          , slice(None)), True ),   # np.transpose(elemNodes[:    , 0    , :    ]), # noqa: E262, E501
                        'x+': ((order      , slice(None), slice(None)), True ),   # np.transpose(elemNodes[order, :    , :    ]), # noqa: E262, E501
                        'x-': ((0          , slice(None), slice(None)), False),   #              elemNodes[0    , :    , :    ],  # noqa: E262, E501
                        'z+': ((slice(None), slice(None), order      ), True )},  # np.transpose(elemNodes[:    , :    , order])},# noqa: E262, E501
                   # Hexahedron
                   8: { 'z-': ((slice(None), slice(None), 0          ), False),   #              elemNodes[:    , :    , 0    ],  # noqa: E262, E501
                        'y-': ((slice(None), 0          , slice(None)), True ),   # np.transpose(elemNodes[:    , 0    , :    ]), # noqa: E262, E501
                        'x+': ((order      , slice(None), slice(None)), True ),   # np.transpose(elemNodes[order, :    , :    ]), # noqa: E262, E501
                        'y+': ((slice(None), order      , slice(None)), False),   #              elemNodes[:    , order, :    ],  # noqa: E262, E501
                        'x-': ((0          , slice(None), slice(None)), False),   #              elemNodes[0    , :    , :    ],  # noqa: E262, E501
                        'z+': ((slice(None), slice(None), order      ), True )}   # np.transpose(elemNodes[:    , :    , order])} # noqa: E262, E501
                 }
    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_cgns: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][dir]
    except KeyError:
        raise KeyError(f'Error in face_to_cgns: face {dir} is not supported')


# > Not cacheable, we pass mesh[meshio.Mesh]
def count_elems(mesh: meshio.Mesh) -> int:
    nElems = 0
    for _, elemType in enumerate(mesh.cells_dict.keys()):
        # Only consider three-dimensional types
        if not any(s in elemType for s in elemTypeClass.type.keys()):
            continue

        ioelems = mesh.get_cells_type(elemType)
        nElems += ioelems.shape[0]
    return nElems


# > Not cacheable, we pass mesh[meshio.Mesh]
def calc_elem_bary(elems: list) -> np.ndarray:
    """
    Compute barycenters of all three-dimensional elements in the mesh.

    Returns:
        elem_bary (np.ndarray): Array of barycenters for all 3D elements, concatenated.
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import numpy as np
    # ------------------------------------------------------

    elem_bary = []
    for elem in elems:
        # Calculate barycenters
        bary = np.mean(mesh_vars.mesh.points[elem.nodes], axis=0)
        elem_bary.append(bary)

    return np.asarray(elem_bary)


@cache
def LINTEN(elemType: int, order: int = 1) -> tuple[np.ndarray, dict[np.int64, int]]:
    """ MESHIO -> IJK ordering for element volume nodes
    """
    # Local imports ----------------------------------------
    # from pyhope.io.formats.cgns import genHEXMAPCGNS
    # from pyhope.io.formats.vtk import genHEXMAPVTK
    from pyhope.io.formats.meshio import TETRMAPMESHIO, PYRAMAPMESHIO, PRISMAPMESHIO, HEXMAPMESHIO
    # ------------------------------------------------------
    # Check if we try to access a curved element with a straight-sided mapping
    if order > 1 and elemType < 200:
        raise ValueError(f'Error in LINTEN: order {order} is not supported for elemType {elemType}')

    match elemType:
        # Straight-sided elements, hard-coded
        case 104:  # Tetraeder
            # return np.array((0, 1, 2, 3))
            TETRTEN = np.array((0, 1, 2, 3))
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENTETR   = {k: v for v, k in enumerate(TETRTEN)}
            return TETRTEN, TENTETR
        case 105:  # Pyramid
            # return np.array((0, 1, 3, 2, 4))
            PYRATEN = np.array((0, 1, 3, 2, 4))
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENPYRA   = {k: v for v, k in enumerate(PYRATEN)}
            return PYRATEN, TENPYRA
        case 106:  # Prism
            # return np.array((0, 1, 2, 3, 4, 5))
            PRISTEN = np.array((0, 1, 2, 3, 4, 5))
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENPRIS   = {k: v for v, k in enumerate(PRISTEN)}
            return PRISTEN, TENPRIS
        case 108:  # Hexaeder
            # return np.array((0, 1, 3, 2, 4, 5, 7, 6))
            HEXTEN = np.array((0, 1, 3, 2, 4, 5, 7, 6))
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENHEX    = {k: v for v, k in enumerate(HEXTEN)}
            return HEXTEN, TENHEX
        # Curved elements, use mapping
        case 204:  # Tetraeder
            _, TETRTEN = TETRMAPMESHIO(order+1)
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENTETR   = {k: v for v, k in enumerate(TETRTEN)}
            return TETRTEN, TENTETR
        case 205:  # Pyramid
            _, PYRATEN = PYRAMAPMESHIO(order+1)
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENPYRA   = {k: v for v, k in enumerate(PYRATEN)}
            return PYRATEN, TENPYRA
        case 206:  # Prism
            _, PRISTEN = PRISMAPMESHIO(order+1)
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENPRIS   = {k: v for v, k in enumerate(PRISTEN)}
            return PRISTEN, TENPRIS
        case 208:  # Hexaeder
            # > HEXTEN : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (1D, tensor-product style)
            # > HEXMAP : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (3D mapping)

            # # CGNS
            # _, HEXTEN = HEXMAPCGNS(order+1)

            # # VTK
            # _, HEXTEN = HEXMAPVTK(order+1)

            # MESHIO
            _, HEXTEN = HEXMAPMESHIO(order+1)
            # meshio accesses them in their own ordering
            # > need to reverse the mapping
            TENHEX    = {k: v for v, k in enumerate(HEXTEN)}
            return HEXTEN, TENHEX
        case _:  # Default
            print('Error in LINTEN, unknown elemType')
            sys.exit(1)


@cache
def LINMAP(elemType: int, order: int = 1) -> npt.NDArray[np.int32]:
    """ MESHIO -> IJK ordering for element corner nodes
    """
    # Local imports ----------------------------------------
    # from pyhope.io.formats.cgns import HEXMAPCGNS
    # from pyhope.io.formats.vtk import HEXMAPVTK
    from pyhope.io.formats.meshio import TETRMAPMESHIO, PYRAMAPMESHIO, PRISMAPMESHIO, HEXMAPMESHIO
    # ------------------------------------------------------
    # Check if we try to access a curved element with a straight-sided mapping
    if order > 1 and elemType < 200:
        raise ValueError(f'Error in LINTEN: order {order} is not supported for elemType {elemType}')

    match elemType:
        # Straight-sided elements, hard-coded
        case 104:  # Tetraeder
            linmap = np.full((2, 2, 2), -1, dtype=np.int32)
            indices = [ (0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1)]
            for i, index in enumerate(indices):
                linmap[index] = i
            return linmap
        case 105:  # Pyramid
            linmap = np.full((2, 2, 2), -1, dtype=np.int32)
            indices = [ (0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0),
                        (0, 0, 1)]
            for i, index in enumerate(indices):
                linmap[index] = i
            return linmap
        case 106:  # Prism
            linmap = np.full((2, 2, 2), -1, dtype=np.int32)
            indices = [ (0, 0, 0), (1, 0, 0), (0, 1, 0),
                        (0, 0, 1), (1, 0, 1), (0, 1, 1)]
            for i, index in enumerate(indices):
                linmap[index] = i
            return linmap
        case 108:  # Hexaeder
            linmap = np.zeros((2, 2, 2), dtype=np.int32)
            indices = [ (0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0),
                        (0, 0, 1), (1, 0, 1), (1, 1, 1), (0, 1, 1) ]
            for i, index in enumerate(indices):
                linmap[index] = i
            return linmap

        # Curved elements, use mapping
        case 204:  # Tetraeder
            TETRMAP, _ = TETRMAPMESHIO(order+1)
            return TETRMAP
        case 205:  # Pyramid
            PYRAMAP, _ = PYRAMAPMESHIO(order+1)
            return PYRAMAP
        case 206:  # Prism
            PRISMAP, _ = PRISMAPMESHIO(order+1)
            return PRISMAP
        case 208:  # Hexaeder
            # > HEXTEN : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (1D, tensor-product style)
            # > HEXMAP : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (3D mapping)

            # # CGNS
            # HEXMAP  , _ = HEXMAPCGNS(order+1)

            # # VTK
            # HEXMAP  , _ = HEXMAPVTK(order+1)

            # MESHIO
            HEXMAP  , _ = HEXMAPMESHIO(order+1)
            return HEXMAP
        case _:  # Default
            print('Error in LINMAP, unknown elemType')
            sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import re
import sys
from typing import Final, Optional
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.mesh.mesh_vars as mesh_vars
from pyhope.mesh.mesh_common import LINMAP
from pyhope.mesh.mesh_common import dir_to_nodes, faces
# ==================================================================================================================================


def check_orientation(ionodes : np.ndarray,
                      elemType: int) -> tuple[bool, Optional[str]]:
    """ Check the orientation of the surface normals
    """
    mapLin   = LINMAP(elemType, order=mesh_vars.nGeo)
    nodes    = ionodes[mapLin]
    iopoints = mesh_vars.mesh.points
    points   = iopoints[ionodes[mapLin]]

    # Center of element
    cElem = np.mean(points, axis=(0, 1, 2))

    success = True
    sface   = None
    for face in faces(elemType):
        # Center of face
        indices, doTransp = dir_to_nodes(face, elemType, mesh_vars.nGeo)
        fnodes  = nodes[indices]
        if doTransp:
            fnodes = fnodes.transpose()
        fpoints = iopoints[fnodes]
        # cFace  = fpoints.mean(axis=tuple(range(fpoints.ndim - 1)))
        cFace  = np.mean(fpoints, axis=(0, 1))

        # Tangent and normal vectors
        nVecFace = cElem - cFace
        # nVecFace = nVecFace / np.linalg.norm(nVecFace)
        nVecFace = nVecFace / np.sqrt(np.dot(nVecFace, nVecFace))
        vec1 = fpoints[-1, 0, :] - fpoints[0, 0, :]
        vec2 = fpoints[0, -1, :] - fpoints[0, 0, :]
        normal = np.cross(vec1, vec2)
        # normal /= np.linalg.norm(normal)
        normal /= np.sqrt(np.dot(normal, normal))

        # Dot product and check if normal points outwards
        dotprod = np.dot(nVecFace, normal)
        if dotprod < 0:
            success = False
            sface   = face
            break
    return success, sface


def process_chunk(chunk) -> np.ndarray:
    """Process a chunk of elements by checking surface normal orientation
    """
    chunk_results = np.fromiter(((check_orientation(ionodes, elemType), iElem)
                                  for iElem, ionodes, elemType in chunk), dtype=object)
    return chunk_results


def OrientMesh() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common_parallel import run_in_parallel
    from pyhope.common.common_vars import np_mtp
    from pyhope.readintools.readintools import GetLogical
    # ------------------------------------------------------

    hopout.routine('Ensuring normals point outward')
    hopout.sep()

    checkSurfaceNormals = GetLogical('CheckSurfaceNormals')
    if not checkSurfaceNormals:
        return None

    mesh = mesh_vars.mesh

    elemNames: Final[dict] = mesh_vars.ELEMTYPE.name
    elemKeys : Final       = mesh_vars.ELEMTYPE.type.keys()
    nElems      = 0
    passedTypes = []

    for elemType in mesh.cells_dict.keys():
        # Only consider three-dimensional types
        if not any(s in elemType for s in elemKeys):
            continue

        # Only consider hexahedrons
        if 'hexahedron' not in elemType:
            passedTypes.append(elemType)
            continue

        # Get the elements
        ioelems  = mesh.get_cells_type(elemType)
        nIOElems = ioelems.shape[0]

        if isinstance(elemType, str):
            elemType = elemNames[elemType]

        # Prepare elements for parallel processing
        if np_mtp > 0:
            tasks = tuple((iElem, ioelems[iElem - nElems], elemType)
                           for iElem in range(nElems, nElems + nIOElems))
            # Run in parallel with a chunk size
            # > Dispatch the tasks to the workers, minimum 10 tasks per worker, maximum 1000 tasks per worker
            res   = run_in_parallel(process_chunk, tasks, chunk_size=max(1, min(1000, max(10, int(len(tasks)/(40.*np_mtp))))))
        else:
            res   = np.fromiter(((check_orientation(ioelems[iElem - nElems], elemType), iElem)
                                  for iElem in range(nElems, nElems + nIOElems)), dtype=object)

        if not np.all([success for (success, _), _ in res]):
            failed_elems = [(iElem + 1, face) for (success, face), iElem in res if not success]
            for iElem, face in failed_elems:
                print(hopout.warn(f'> Element {iElem}, Side {face}'))
            sys.exit(1)

        # Add to nElems
        nElems += nIOElems

    # Warn if we passed any element types
    if len(passedTypes) > 0:
        print(hopout.warn('Ignored element type{}: {}'.format('s' if len(passedTypes) > 1 else '',
                                                              [re.sub(r"\d+$", "", s) for s in passedTypes])))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import time
# from sortedcontainers import SortedDict
from collections import defaultdict
from typing import Final, Tuple
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def time_function(func, *args, **kwargs) -> float:
    """ A helper function to measure the execution time of an arbitrary function.

    Parameters:
    func (callable): The function to be timed.
    *args: Positional arguments to pass to the function.
    **kwargs: Keyword arguments to pass to the function.

    Returns:
    The return value of the function being timed.
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    # ------------------------------------------------------
    tStart: Final[float] = time.time()
    result               = func(*args, **kwargs)
    tEnd:   Final[float] = time.time()
    tFunc:  Final[float] = tEnd - tStart
    hopout.info(  hopout.Colors.BANNERA + f'Function {func.__name__} required {tFunc:.6f} seconds to complete.'
                + hopout.Colors.END)

    return result


def allocate_or_resize( dict: dict, key: str, shape: Tuple[int, int]) -> Tuple[dict, int]:
    """ Allocate or resize a numpy array in a dictionary.
    """
    offset = 0
    if key not in dict:
        dict[key] = np.ndarray(shape, dtype=np.uint)
    else:
        offset = dict[key].shape[0]
        new_len = offset + shape[0]
        dict[key] = np.resize(dict[key],  (new_len, shape[1]))

    return dict, offset


class IndexedLists:
    def __init__(self) -> None:
        # PERF: The unsorted dict is faster for adding and removing elements
        # Create a SortedDict to keep the data sorted by index
        # self.data = SortedDict()

        # Use a plain dict for faster key operations
        self.data = {}
        # Inverse mapping: for each value, store the set of keys (indices) that contain it
        self._inverse = defaultdict(set)

    def add(self, index: int, values) -> None:
        """ Add a sublist at a specific integer index
        """
        value_set = set(values)  # Use set for fast removals
        self.data[index] = value_set
        for v in value_set:
            self._inverse[v].add(index)

    def remove_index(self, indices) -> None:
        """ Remove the sublist at idx and remove the integer idx from all remaining sublists
        """
        if isinstance(indices, int):
            # Convert to a set for fast operations
            indices = {indices}
        else:
            # Convert list to set for O(1) lookups
            indices = set(indices)

        # Create a set to hold all affected keys
        affected_keys = set()
        # Remove keys directly and collect affected keys from inverse mapping
        for idx in indices:
            if idx in self.data:
                # Remove the entire key; also clean up the inverse mapping
                for v in self.data[idx]:
                    self._inverse[v].discard(idx)
                del self.data[idx]
            affected_keys |= self._inverse.pop(idx, set())

        # Now update each affected key only once while checking for emptiness
        for key in affected_keys:
            if key in self.data:
                self.data[key].difference_update(indices)

    def __getitem__(self, index) -> list[int]:
        """ Retrieve a sublist by index
        """
        return list(self.data[index])  # Convert set back to list when accessed

    def __repr__(self) -> str:
        return repr({k: list(v) for k, v in self.data.items()})  # Convert to list for cleaner output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def legendre_gauss_lobatto_nodes(order: int) -> tuple[np.ndarray, np.ndarray]:
    """ Return Legendre-Gauss-Lobatto nodes and weights for a given order in 1D
    """
    order -= 1
    # Special cases for small N
    if order == 1:
        return np.array((-1, 1)), np.array((1, 1))

    # Compute the initial guess for the LGL nodes (roots of P'_N)
    nodes = np.cos(np.pi * np.arange(order+1) / order)

    # Initialize the Legendre polynomial and its derivative
    p = np.zeros((order+1, order+1))

    # Iteratively solve for the LGL nodes using Newton's method
    xOld = 2 * np.ones_like(nodes)
    tol = 1e-14
    while np.max(np.abs(nodes - xOld)) > tol:
        xOld = nodes.copy()
        p[:, 0] = 1
        p[:, 1] = nodes
        for k in range(2, order+1):
            p[:, k] = ((2*k-1) * nodes * p[:, k-1] - (k-1) * p[:, k-2]) / k
        nodes -= (nodes * p[:, order] - p[:, order-1]) / (order * (p[:, order]))

    # The LGL nodes
    nodes = np.sort(nodes)

    # Compute the LGL weights
    weights = 2 / (order * (order + 1) * (p[:, order]**2))

    return nodes, weights


def legendre_gauss_nodes(order: int) -> tuple[np.ndarray, np.ndarray]:
    """ Return Legendre-Gauss nodes and weights for a given order in 1D
    """
    nodes, weights = np.polynomial.legendre.leggauss(order)
    return nodes, weights


def barycentric_weights(_: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the barycentric weights for a given node set
        > Algorithm 30, Kopriva
    """
    # Create a difference matrix (x_i - x_j) for all i, j
    diff_matrix = xGP[:, np.newaxis] - xGP[np.newaxis, :]

    # Set the diagonal to 1 to avoid division by zero (diagonal elements will not be used)
    np.fill_diagonal(diff_matrix, 1.0)

    # Compute the product of all differences for each row (excluding the diagonal)
    wBary = np.prod(diff_matrix, axis=1)

    # Take the reciprocal to get the final barycentric weights
    wBary = 1.0 / wBary

    return wBary


def polynomial_derivative_matrix(order: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the polynomial derivative matrix for a given node set
        > Algorithm 37, Kopriva
    """
    wBary = barycentric_weights(order, xGP)
    D     = np.zeros((order, order), dtype=float)

    for iLagrange in range(order):
        for iGP in range(order):
            if iLagrange != iGP:
                D[iGP, iLagrange] = wBary[iLagrange]/(wBary[iGP]*(xGP[iGP]-xGP[iLagrange]))
                D[iGP, iGP      ] = D[iGP, iGP] - D[iGP, iLagrange]

    return D


def lagrange_interpolation_polys(x: float, order: int, xGP: np.ndarray, wBary: np.ndarray) -> np.ndarray:
    """ Computes all Lagrange functions evaluated at position x in [-1;1]
        > Algorithm 34, Kopriva
    """
    # Equal points need special treatment
    lagrange = np.zeros(order)
    for iGP in range(order):
        if abs(x - xGP[iGP]) < 1.E-14:
            lagrange[iGP] = 1
            return lagrange

    tmp = 0.
    for iGP in range(order):
        lagrange[iGP] = wBary[iGP] / (x-xGP[iGP])
        tmp   += lagrange[iGP]

    # Normalize
    lagrange = lagrange/tmp
    return lagrange


def calc_vandermonde(n_In: int, n_Out: int, wBary_In: np.ndarray, xi_In: np.ndarray, xi_Out: np.ndarray) -> np.ndarray:
    """ Build a 1D Vandermonde matrix using the Lagrange basis functions of degree N_In,
        evaluated at the interpolation points xi_Out
    """
    Vdm = np.zeros((n_Out, n_In))
    for iXI in range(n_Out):
        Vdm[iXI, :] = lagrange_interpolation_polys(xi_Out[iXI], n_In, xi_In, wBary_In)
    return Vdm


# def change_basis_3D(dim1: int, n_In: int, n_Out: int, Vdm: np.ndarray, x3D_In: np.ndarray) -> np.ndarray:
def change_basis_3D(Vdm: np.ndarray, x3D_In: np.ndarray) -> np.ndarray:
    """ Interpolate a 3D tensor product Lagrange basis defined by (N_in+1) 1D interpolation point positions xi_In(0:N_In)
        to another 3D tensor product node positions (number of nodes N_out+1)
        defined by (N_out+1) interpolation point  positions xi_Out(0:N_Out)
        xi is defined in the 1DrefElem xi=[-1,1]
    """
    # First contraction along the iN_In axis (axis 1 of Vdm, axis 1 of x3D_In)
    x3D_Buf1 = np.tensordot(Vdm, x3D_In, axes=(1, 1))
    x3D_Buf1 = np.moveaxis(x3D_Buf1, 0, 1)  # Correct the shape to (dim1, n_Out, n_In, n_In)

    # Second contraction along the jN_In axis (axis 1 of Vdm, axis 2 of x3D_Buf1)
    x3D_Buf2 = np.tensordot(Vdm, x3D_Buf1, axes=(1, 2))
    x3D_Buf2 = np.moveaxis(x3D_Buf2, 0, 2)  # Correct the shape to  (dim1, n_Out, n_Out, n_In)

    # Third contraction along the kN_In axis (axis 1 of Vdm, axis 3 of x3D_Buf2)
    x3D_Out  = np.tensordot(Vdm, x3D_Buf2, axes=(1, 3))
    x3D_Out  = np.moveaxis(x3D_Out , 0, 3)  # Correct the shape to (dim1, n_Out, n_Out, n_Out)

    return x3D_Out


def change_basis_2D(Vdm: np.ndarray, x2D_In: np.ndarray) -> np.ndarray:
    """ Interpolate a 2D tensor product Lagrange basis defined by (N_in+1) 1D interpolation point positions xi_In(0:N_In)
        to another 2D tensor product node positions (number of nodes N_out+1)
        defined by (N_out+1) interpolation point positions xi_Out(0:N_Out)
        xi is defined in the 1D reference element xi=[-1,1]
    """
    # First contraction along the iN_In axis (axis 1 of Vdm, axis 1 of x2D_In)
    x2D_Buf1 = np.tensordot(Vdm, x2D_In, axes=(1, 1))
    x2D_Buf1 = np.moveaxis(x2D_Buf1, 0, 1)  # Correct the shape to (dim1, n_Out, n_In, n_In)

    # Second contraction along the jN_In axis (axis 1 of Vdm, axis 2 of x2D_Buf1)
    x2D_Out = np.tensordot(Vdm, x2D_Buf1, axes=(1, 2))
    x2D_Out = np.moveaxis(x2D_Out, 0, 2)  # Correct the shape to  (dim1, n_Out, n_Out, n_In)

    return x2D_Out


def evaluate_jacobian(xGeo_In: np.ndarray, VdmGLtoAP: np.ndarray, D_EqToGL: np.ndarray) -> np.ndarray:
    # Perform tensor contraction for the first derivative (Xi direction)
    dXdXiGL   = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 1))
    dXdXiGL   = np.moveaxis(dXdXiGL  , 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)

    # Perform tensor contraction for the second derivative (Eta direction)
    dXdEtaGL  = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 2))
    dXdEtaGL  = np.moveaxis(dXdEtaGL , 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)

    # Perform tensor contraction for the third derivative (Zeta direction)
    dXdZetaGL = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 3))
    dXdZetaGL = np.moveaxis(dXdZetaGL, 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)

    # Change basis for each direction
    dXdXiAP   = change_basis_3D(VdmGLtoAP, dXdXiGL  )
    dXdEtaAP  = change_basis_3D(VdmGLtoAP, dXdEtaGL )
    dXdZetaAP = change_basis_3D(VdmGLtoAP, dXdZetaGL)

    # Precompute cross products between dXdEtaAP and dXdZetaAP for all points
    cross_eta_zeta = np.cross(dXdEtaAP, dXdZetaAP, axis=0)  # Shape: (3, nGeoRef, nGeoRef, nGeoRef)

    # Fill output Jacobian array
    jacOut = np.einsum('ijkl,ijkl->jkl', dXdXiAP, cross_eta_zeta)

    return jacOut


# INFO: ALTERNATIVE VERSION, CACHING VDM, D
# class JacobianEvaluator:
#     def __init__(self, VdmGLtoAP: np.ndarray, D_EqToGL: np.ndarray) -> None:
#         self.VdmGLtoAP: Final[np.ndarray] = VdmGLtoAP
#         self.D_EqToGL:  Final[np.ndarray] = D_EqToGL
#
#     def evaluate_jacobian(self, xGeo_In: np.ndarray) -> np.ndarray:
#         # Perform tensor contraction for the first derivative (Xi direction)
#         dXdXiGL   = np.tensordot(self.D_EqToGL, xGeo_In, axes=(1, 1))
#         dXdXiGL   = np.moveaxis(dXdXiGL  , 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Perform tensor contraction for the second derivative (Eta direction)
#         dXdEtaGL  = np.tensordot(self.D_EqToGL, xGeo_In, axes=(1, 2))
#         dXdEtaGL  = np.moveaxis(dXdEtaGL , 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Perform tensor contraction for the third derivative (Zeta direction)
#         dXdZetaGL = np.tensordot(self.D_EqToGL, xGeo_In, axes=(1, 3))
#         dXdZetaGL = np.moveaxis(dXdZetaGL, 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Change basis for each direction
#         dXdXiAP   = change_basis_3D(self.VdmGLtoAP, dXdXiGL  )
#         dXdEtaAP  = change_basis_3D(self.VdmGLtoAP, dXdEtaGL )
#         dXdZetaAP = change_basis_3D(self.VdmGLtoAP, dXdZetaGL)
#
#         # Precompute cross products between dXdEtaAP and dXdZetaAP for all points
#         cross_eta_zeta = np.cross(dXdEtaAP, dXdZetaAP, axis=0)  # Shape: (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Fill output Jacobian array
#         jacOut = np.einsum('ijkl,ijkl->jkl', dXdXiAP, cross_eta_zeta)
#
#         return jacOut



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import copy
import gc
import itertools
import os
import shutil
import sys
import tempfile
# from dataclasses import dataclass, field
from functools import cache
from string import digits
from typing import cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import h5py
import meshio
import numpy as np
from alive_progress import alive_bar
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


@cache
def NDOFperElemType(elemType: str, nGeo: int) -> int:
    """ Calculate the number of degrees of freedom for a given element type
    """
    match elemType:
        case _ if elemType.startswith('triangle'):
            return round((nGeo+1)*(nGeo+2)/2.)
        case _ if elemType.startswith('quad'):
            return round((nGeo+1)**2)
        case _ if elemType.startswith('tetra'):
            return round((nGeo+1)*(nGeo+2)*(nGeo+3)/6.)
        case _ if elemType.startswith('pyramid'):
            return round((nGeo+1)*(nGeo+2)*(2*nGeo+3)/6.)
        case _ if elemType.startswith('wedge'):
            return round((nGeo+1)**2 *(nGeo+2)/2.)
        case _ if elemType.startswith('hexahedron'):
            return round((nGeo+1)**3)
        case _:
            raise ValueError(f'Unknown element type {elemType}')


# @dataclass
# class FaceOrdering:
#     side_type: str
#     nGeo     : int
#     order    : np.ndarray = field(init=False)
#
#     def __post_init__(self):
#         self.order = self.compute_ordering()
#
#     def compute_ordering(self) -> np.ndarray:
@cache
def FaceOrdering(side_type: str, nGeo: int) -> np.ndarray:
    """
    Compute the permutation ordering to convert from tensor-product ordering
    to meshio ordering for a face of a given type ('quad' or 'triangle')
    and polynomial order nGeo.

    For quadrilaterals, total nodes = (nGeo+1)**2.
      - For nGeo==1, the natural ordering is [0, 1, 2, 3].
      - For nGeo>1, the ordering is:
          * Corners: bottom-left, bottom-right, top-right, top-left;
          * Then the bottom edge (excluding corners, left-to-right);
          * Then the right  edge (excluding corners, bottom-to-top);
          * Then the top    edge (excluding corners, right-to-left);
          * Then the left   edge (excluding corners, top-to-bottom);
          * Finally, the interior nodes in row-major order.

    For triangles, total nodes = (nGeo+1)*(nGeo+2)//2.
      - For nGeo==1, the natural ordering is [0, 1, 2].
      - For nGeo>1, we generate the tensor ordering as all (i,j) pairs
        with i+j <= nGeo (in lexicographical order) and then reorder so that:
          * Vertices come first: (0,0), (nGeo,0), (0,nGeo);
          * Followed by edge nodes (in order along each edge);
          * And then the interior nodes in their natural order.
    """
    if side_type.lower() == 'quad':
        # Total nodes on face: (nGeo+1)**2
        if nGeo == 1:
            return np.arange(4)
        else:
            n           = nGeo
            grid        = np.arange((n+1)**2).reshape(n+1, n+1)
            # Corners: bottom-left, bottom-right, top-right, top-left
            corners     = np.array((grid[0, 0], grid[0, n], grid[n, n], grid[n, 0]))
            # Bottom edge (excluding corners): row 0, columns 1 to n-1 (left-to-right)
            bottom_edge = grid[0, 1:n]
            # Right edge: column n, rows 1 to n-1 (bottom-to-top)
            right_edge  = grid[1:n, n]
            # Top edge: row n, columns n-1 to 1 (right-to-left)
            top_edge    = grid[n, n-1:0:-1]
            # Left edge: column 0, rows n-1 to 1 (top-to-bottom)
            left_edge   = grid[n-1:0:-1, 0]
            # Interior nodes: remaining nodes in row-major order
            interior    = grid[1:n, 1:n].flatten()
            # Assemble ordering: corners, edges, interior
            # order       = np.concatenate((corners, bottom_edge, right_edge, top_edge, left_edge, interior))
            order       = np.concatenate((corners, bottom_edge, right_edge, top_edge, left_edge, interior))
            return order

    elif side_type.lower() == 'triangle':
        # Total nodes on face: (nGeo+1)*(nGeo+2)//2
        if nGeo == 1:
            return np.arange(3)
        else:
            p           = nGeo
            # Build the tensor ordering as a list of (i, j) for which i+j <= p.
            nodes       = []
            for i in range(p+1):
                for j in range(p+1 - i):
                    nodes.append((i, j))
            # Define vertices in the reference triangle:
            vertices    = [(0, 0), (p, 0), (0, p)]
            # Edge from vertex0 (0,0) to vertex1 (p,0): nodes with j==0 (excluding vertices)
            edge01      = [(i, 0) for i in range(1, p)]
            # Edge from vertex1 (p,0) to vertex2 (0,p): nodes on the line i+j==p (excluding vertices)
            edge12      = [(i, p-i) for i in range(p-1, 0, -1)]
            # Edge from vertex2 (0,p) to vertex0 (0,0): nodes with i==0 (excluding vertices)
            edge20      = [(0, j) for j in range(1, p)]
            # Interior nodes: those not on the boundary
            boundary    = set(vertices + edge01 + edge12 + edge20)
            interior    = [node for node in nodes if node not in boundary]
            # Assemble ordering: vertices, then edge nodes in order, then interior nodes.
            desired     = vertices + edge01 + edge12 + edge20 + interior
            order       = [nodes.index(nd) for nd in desired]
            return np.array(order)
    else:
        raise ValueError(f'Unsupported side type: {side_type}')


def ReadHOPR(fnames: list, mesh: meshio.Mesh) -> meshio.Mesh:
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.basis.basis_basis import barycentric_weights, calc_vandermonde, change_basis_3D
    from pyhope.mesh.mesh_common import LINTEN
    from pyhope.mesh.mesh_common import faces, face_to_nodes
    from pyhope.mesh.mesh_vars import ELEMTYPE
    # ------------------------------------------------------

    hopout.sep()

    # Create an empty meshio object
    points   = mesh.points if len(mesh.points.shape)>1 else np.zeros((0, 3), dtype=np.float64)
    pointl   = cast(list, points.tolist())
    cells    = mesh.cells_dict
    cellsets = {}

    nodeCoords   = mesh.points
    offsetnNodes = nodeCoords.shape[0]
    nSides       = np.zeros(2, dtype=int)

    # Vandermonde for changeBasis
    VdmEqHdf5ToEqMesh = np.array(())
    mortarTypeToSkip  = {1: 4, 2: 2, 3: 2}

    # Instantiate ELEMTYPE
    elemTypeClass = ELEMTYPE()

    for fname in fnames:
        # Check if the file is using HDF5 format internally
        if not h5py.is_hdf5(fname):
            hopout.warning('[󰇘]/{} is not in HDF5 format, exiting...'.format(os.path.basename(fname)))
            sys.exit(1)

        # Create a temporary directory and keep it existing until manually cleaned
        tfile = tempfile.NamedTemporaryFile(delete=False)
        tname = tfile.name
        # Alternatively, load the file directly into tmpfs for faster access
        shutil.copyfile(fname, tname)

        with h5py.File(tname, mode='r') as f:
            # Check if file contains the Hopr version
            if 'HoprVersion' not in f.attrs:
                hopout.warning('[󰇘]/{} does not contain the Hopr version, exiting...'.format(os.path.basename(fname)))
                sys.exit(1)

            # Read the globalNodeIDs
            nodeInfo   = np.array(f['GlobalNodeIDs'])

            # Read the nodeCoords
            nodeCoords = np.array(f['NodeCoords'])

            # Read nGeo
            nGeo       = int(cast(int, f.attrs['Ngeo']))

            # Try reading in periodic vector if it is not provided in file
            if len(mesh_vars.vvs) == 0:
                try:
                    vvs = np.array(f['VV'])
                    mesh_vars.vvs = [dict() for _ in range(vvs.shape[0])]
                    for iVV, _ in enumerate(mesh_vars.vvs):
                        mesh_vars.vvs[iVV] = dict()
                        mesh_vars.vvs[iVV]['Dir'] = vvs[iVV]
                    # Output vectors
                    print(hopout.warn('Periodicity vectors not defined in parameter file. '
                                      'Reading the vectors from given PyHOPE mesh!'))
                    hopout.sep()
                    hopout.routine('The following vectors were found:')
                    for iVV, vv in enumerate(mesh_vars.vvs):
                        hopout.printoption('vv[{}]'.format(iVV+1),'{0:}'.format(np.round(vv['Dir'],6)), 'READ IN')
                    hopout.sep()
                # old hopr files might not contain the VV
                except KeyError:
                    pass

            if nGeo == mesh_vars.nGeo:
                # only retain the unique nodes
                indices    = np.unique(nodeInfo, return_index=True)[1]
                nodeCoords = nodeCoords[indices]
                # points     = np.append(points, nodeCoords, axis=0)
                # IMPORTANT: We need to extend the list of points, not append to it
                pointl.extend(nodeCoords.tolist())
            else:
                # ChangeBasis on the non-unique nodes
                # > Currently only supported for hexahedrons
                filename = os.path.basename(fname)
                print(hopout.warn(f'[󰇘]/{filename} has different polynomial order than the current mesh, converting...',
                      length=999))
                print(hopout.warn(f'> NGeo [HDF5] = {nGeo}, NGeo [Mesh] = {mesh_vars.nGeo}') + '\n')

                # Compute the equidistant point set used by HOPR
                xEqHdf5     = np.linspace(-1, 1, num=nGeo+1, dtype=np.float64)
                wBaryEqHdf5 = barycentric_weights(nGeo, xEqHdf5)

                # Compute the equidistant point set used by meshIO
                xEqMesh     = np.linspace(-1, 1, num=mesh_vars.nGeo+1, dtype=np.float64)
                # wBaryEqMesh = barycentric_weights(mesh_vars.nGeo, xEqMesh)

                # Compute the Vandermonde matrix
                VdmEqHdf5ToEqMesh = calc_vandermonde(nGeo+1, mesh_vars.nGeo+1, wBaryEqHdf5, xEqHdf5, xEqMesh)

            # Read the elemInfo and sideInfo
            elemInfo   = np.array(f['ElemInfo'])
            sideInfo   = np.array(f['SideInfo'])
            BCNames    = [s.strip().decode('utf-8') for s in cast(h5py.Dataset, f['BCNames'])]

            # Cache the mapping here, so we consider the mesh order
            linCache   = {}

            with alive_bar(len(elemInfo), title='│             Processing Elements', length=33) as bar:
                # Construct the elements, meshio format
                for elem in elemInfo:
                    # Correct ElemType if NGeo is changed
                    elemNum  = elem[0] % 100
                    elemNum += 200 if mesh_vars.nGeo > 1 else 100

                    # Obtain the element type
                    elemType = elemTypeClass.inam[elemNum]
                    if len(elemType) > 1:
                        elemType  = elemType[0].rstrip(digits)
                        elemType += str(NDOFperElemType(elemType, mesh_vars.nGeo))
                    else:
                        elemType  = elemType[0]

                    # ChangeBasis currently only supported for hexahedrons
                    if elemNum in linCache:
                        mapLin = linCache[elemNum]
                    else:
                        _, mapLin = LINTEN(elemNum, order=mesh_vars.nGeo)
                        mapLin    = np.array(tuple(mapLin[np.int64(i)] for i in range(len(mapLin))))
                        linCache[elemNum] = mapLin

                    if nGeo == mesh_vars.nGeo:
                        elemIDs   = np.arange(elem[4], elem[5])
                        elemNodes = elemIDs[mapLin[:len(elemIDs)]]
                        elemNodes = np.expand_dims(nodeInfo[elemNodes] - 1 + offsetnNodes, axis=0)
                    else:
                        nElemNode = (mesh_vars.nGeo+1)**3
                        # elemIDs   = np.arange(points.shape[0], points.shape[0]+nElemNode, dtype=np.uint64)
                        elemIDs   = np.arange(len(pointl), len(pointl)+nElemNode, dtype=np.uint64)
                        elemNodes = elemIDs[mapLin[:nElemNode]]
                        # This needs no offset as we already accounted for the number of points in elemIDs
                        elemNodes = np.expand_dims(         elemNodes                    , axis=0)

                        # This is still in tensor-product format
                        meshNodes = nodeCoords[np.arange(elem[4], elem[5])].reshape((nGeo+1, nGeo+1, nGeo+1, 3))
                        meshNodes = meshNodes.transpose(3, 0, 1, 2)
                        try:
                            meshNodes = change_basis_3D(VdmEqHdf5ToEqMesh, meshNodes)
                            meshNodes = meshNodes.transpose(1, 2, 3, 0)
                            meshNodes = meshNodes.reshape((int((mesh_vars.nGeo+1)**3.), 3))
                            # points    = np.append(points, meshNodes, axis=0)
                            # IMPORTANT: We need to extend the list of points, not append to it
                            pointl.extend(meshNodes.tolist())
                        except UnboundLocalError:
                            raise UnboundLocalError('Something went wrong with the change basis')

                    cells.setdefault(elemType, []).append(elemNodes.astype(np.uint64))

                    # Attach the boundary sides
                    sCounter = 0
                    sideRange = iter(range(elem[2], elem[3]))  # Create an iterator for the loop
                    for index in sideRange:
                        # Obtain the side type
                        sideType  = sideInfo[index, 0]
                        nbElemID  = sideInfo[index, 2]
                        sideBC    = sideInfo[index, 4]

                        BCName    = BCNames[sideBC-1].lower()
                        face      = faces(elemType)[sCounter]

                        # Get the number of corners
                        nCorners  = abs(sideType) % 10

                        # Assmble the side information
                        sideNum   = 0      if nCorners == 4 else 1           # noqa: E272
                        sideBase  = 'quad' if nCorners == 4 else 'triangle'  # noqa: E272
                        sideHO    = '' if mesh_vars.nGeo == 1 else str(NDOFperElemType(sideBase, mesh_vars.nGeo))
                        sideName  = sideBase + sideHO

                        # Map the face ordering from tensor-product to meshio
                        order     = FaceOrdering(sideBase, mesh_vars.nGeo)
                        corners   = elemNodes[0][face_to_nodes(face, elemType, mesh_vars.nGeo)]
                        corners   = corners.flatten()[order]
                        sideNodes = np.expand_dims(corners, axis=0)

                        # Add the side to the cells
                        cells.setdefault(sideName, []).append(sideNodes.astype(np.uint64))

                        # Increment the side counter
                        sCounter        += 1
                        nSides[sideNum] += 1

                        # Account for mortar sides
                        if nbElemID < 0:
                            # Side is a big mortar side
                            # > Skip the nVirtualSides small virtual sides
                            nVirtualSides = mortarTypeToSkip[abs(nbElemID)]
                            list(itertools.islice(sideRange, nVirtualSides))

                        if sideBC == 0:
                            continue

                        # Add the side to the cellset
                        # > CS1: We create a dictionary of the BC sides and types that we want
                        cellsets.setdefault(BCName, {}).setdefault(sideName, []).append(nSides[sideNum] - 1)

                    # Update progress bar
                    bar()

                # Update the offset for the next file
                # offsetnNodes = points.shape[0]
                offsetnNodes = len(pointl)

        # Cleanup temporary file
        if tfile is not None:
            os.unlink(tfile.name)

        hopout.sep()

    # After processing all elements, convert each list of arrays to one array
    # > Convert the list of cells to numpy arrays
    for cell_type in cells:
        cells[cell_type] = np.concatenate([a if a.ndim == 2 else a.reshape(1, -1) for a in cells[cell_type]], axis=0)

    # Convert the list of cellsets to numpy arrays
    for bc in cellsets:
        for side in cellsets[bc]:
            cellsets[bc][side] = np.array(cellsets[bc][side], dtype=np.uint64)

    # Convert points_list back to a NumPy array
    points = np.array(pointl)

    # > CS2: We create a meshio.Mesh object without cell_sets
    mesh   = meshio.Mesh(points    = points,    # noqa: E251
                         cells     = cells)     # noqa: E251

    # > CS3: We build the cell sets depending on the cells
    cell_sets  = mesh.cell_sets
    cell_types = [s for s in mesh.cells_dict.keys()]
    cell_list  = [None for _ in cell_types]

    for key, val in cellsets.items():
        for v_key, v_val in val.items():
            if key in cell_sets.keys():
                entry = cell_sets[key]
            else:
                entry = copy.copy(cell_list)

            # Find matching cell type and populate the corresponding entry
            if entry[cell_types.index(v_key)] is not None:
                entry[cell_types.index(v_key)] = np.append(cast(np.ndarray, entry[cell_types.index(v_key)]), v_val)  # type: ignore
            else:
                entry[cell_types.index(v_key)] = v_val

            # Assign the entry to the cell set
            cell_sets[key] = entry  # type: ignore

    # > CS4: We create the final meshio.Mesh object with cell_sets
    mesh   = meshio.Mesh(points    = points,     # noqa: E251
                         cells     = cells,      # noqa: E251
                         cell_sets = cell_sets)  # noqa: E251

    # Run garbage collector to release memory
    gc.collect()

    return mesh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import copy
import gc
import sys
from typing import Final, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
from scipy import spatial
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def EliminateDuplicates() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.mesh.connect.connect import find_bc_index
    # ------------------------------------------------------
    hopout.routine('Removing duplicate points')

    bcs:   Final[list] = mesh_vars.bcs
    vvs:   Final[list] = mesh_vars.vvs

    # Native meshio data
    mesh               = mesh_vars.mesh
    points: np.ndarray = mesh.points
    cells: Final[list] = mesh.cells
    csets: Final[dict] = mesh.cell_sets
    cdict: Final[dict] = mesh.cells_dict

    # Find the mapping to the (N-1)-dim elements
    csetMap = { key: tuple(i for i, cell in enumerate(cset) if cell is not None and cast(np.ndarray, cell).size > 0)
                             for key, cset in csets.items()}

    # Create new periodic nodes per (original node, boundary) pair
    # > Use a dictionary mapping (node, bc_key) --> new node index
    node_bc_translation = {}
    # > Create a list of points to append to the mesh
    pointl  = cast(list, points.tolist())

    for bc_key, cset in csets.items():
        # Find the matching boundary condition
        bcID = find_bc_index(bcs, bc_key)

        # Ignore the volume zones
        if 'Zone' in bc_key:
            continue

        if bcID is None:
            hopout.warning(f'Could not find BC {bc_key} in list, exiting...')
            sys.exit(1)

        # Only process periodic boundaries in the positive direction
        if bcs[bcID].type[0] != 1 or bcs[bcID].type[3] < 0:
            continue

        iVV = bcs[bcID].type[3]
        VV  = vvs[np.abs(iVV)-1]['Dir'] * np.sign(iVV)

        for iMap in csetMap[bc_key]:
            # Only process 2D faces (quad or triangle)
            if not any(s in tuple(cdict)[iMap] for s in ['quad', 'triangle']):
                continue

            iBCsides = np.array(cset[iMap]).astype(int)
            mapFaces = cells[iMap].data

            for iSide in iBCsides:
                for node in mapFaces[iSide]:
                    # Create a unique key for (node, boundary) pair.
                    key_pair = (node, bc_key)

                    # Ignore nodes that have already been processed for this boundary
                    if key_pair in node_bc_translation:
                        continue

                    # Create the new periodic node by applying the boundary's translation.
                    new_node    = points[node] + VV
                    # mesh.points = np.vstack((mesh.points, new_node))
                    pointl.append(new_node)
                    node_bc_translation[key_pair] = len(pointl) - 1

    # Convert the list of points back to an array
    points = np.array(pointl)
    mesh_vars.mesh.points = points
    del pointl

    # At this point, each (node, bc_key) pair has its own new node
    # > Store these in a mapping (here, keys remain as tuples) for later reference
    periNodes = node_bc_translation.copy()

    # Eliminate duplicate points
    points, inverseIndices = np.unique(points, axis=0, return_inverse=True)

    # Update the mesh
    for cell in cells:
        # Map the old indices to the new ones
        # cell.data = np.vectorize(lambda idx: inverseIndices[idx])(cell.data)
        # Efficiently map all indices in one operation
        cell.data = inverseIndices[cell.data]

    # Update periNodes accordingly
    tmpPeriNodes = {}
    for (node, bc_key), new_node in periNodes.items():
        tmpPeriNodes[(inverseIndices[node], bc_key)] = inverseIndices[new_node]
    periNodes = copy.copy(tmpPeriNodes)
    del tmpPeriNodes

    # Also, remove near duplicate points
    # Create a KDTree for the mesh points
    mesh_vars.mesh.points = points
    tree   = spatial.KDTree(points)

    # Filter the valid three-dimensional cell types
    valid_cells = tuple(cell for cell in cells if any(s in cell.type for s in mesh_vars.ELEMTYPE.type.keys()))

    tol = mesh_vars.tolExternal
    bbs = np.empty(len(valid_cells), dtype=float)

    for i, cell in enumerate(valid_cells):
        edata   = np.array(cell.data)
        ecoords = points[edata]
        # Compute the ptp (range) along the vertex axis (axis=1) for each element.
        ptp     = np.ptp(ecoords, axis=1)
        # For each element type, take the minimum across dimensions
        bbs[i]  = ptp.min(axis=1).min()

    # Set the tolerance to 10% of the bounding box of the smallest element
    tol = np.max([tol, bbs.min() / ((mesh_vars.nGeo+1)*10.) ])

    # Find all points within the tolerance
    clusters = tree.query_ball_point(points, r=tol)
    del tree

    # Map each point to its cluster representative (first point in the cluster)
    # > Choose the minimum index as the representative for consistency
    indices = np.fromiter((min(cluster) for cluster in clusters), dtype=int)

    # Eliminate duplicates
    _, inverseIndices = np.unique(indices, return_inverse=True)
    mesh_vars.mesh.points = points[np.unique(indices)]
    del indices

    # Update the mesh cells
    for cell in cells:
        cell.data = inverseIndices[cell.data]

    # Update the periodic nodes
    tmpPeriNodes = {}
    for (node, bc_key), new_node in periNodes.items():
        tmpPeriNodes[(inverseIndices[node], bc_key)] = inverseIndices[new_node]
    mesh_vars.periNodes = tmpPeriNodes

    del inverseIndices

    # Run garbage collector to release memory
    gc.collect()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
from collections import OrderedDict
from typing import Final
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
import h5py
import heapq
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def DefineIO() -> None:
    # Local imports ----------------------------------------
    from pyhope.io.io_vars import MeshFormat
    from pyhope.readintools.readintools import CreateIntFromString, CreateIntOption, CreateLogical, CreateSection, CreateStr
    # ------------------------------------------------------

    CreateSection('Output')
    CreateStr('ProjectName', help='Name of output files')
    CreateIntFromString('OutputFormat', default='HDF5', help='Mesh output format')
    CreateIntOption(    'OutputFormat', number=MeshFormat.FORMAT_HDF5, name='HDF5')
    CreateIntOption(    'OutputFormat', number=MeshFormat.FORMAT_VTK , name='VTK')
    CreateIntOption(    'OutputFormat', number=MeshFormat.FORMAT_GMSH, name='GMSH')
    CreateLogical(      'DebugVisu'   , default=False , help='Launch the GMSH GUI to visualize the mesh')


def InitIO() -> None:
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import GetIntFromStr, GetLogical, GetStr
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('INIT OUTPUT...')

    io_vars.projectname  = GetStr('ProjectName')
    io_vars.outputformat = GetIntFromStr('OutputFormat')

    io_vars.debugvisu    = GetLogical('DebugVisu')

    # hopout.info('INIT OUTPUT DONE!')


def IO() -> None:
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common_vars import Common
    from pyhope.io.io_vars import MeshFormat, ELEM, ELEMTYPE
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('OUTPUT MESH...')

    match io_vars.outputformat:
        case MeshFormat.FORMAT_HDF5:
            mesh  = mesh_vars.mesh
            elems = mesh_vars.elems
            sides = mesh_vars.sides

            nElems = len(elems)
            nSides = len(sides)
            nNodes = sum(s.nodes.size for s in elems)  # number of non-unique nodes

            bcs   = mesh_vars.bcs
            nBCs  = len(bcs)

            pname = io_vars.projectname
            fname = '{}_mesh.h5'.format(pname)

            elemInfo, sideInfo, nodeInfo, nodeCoords, \
            FEMElemInfo, vertexInfo, vertexConnectInfo, edgeInfo, edgeConnectInto, \
            elemCounter = getMeshInfo()

            # Print the final output
            hopout.sep()
            for elemType in ELEM.TYPES:
                if elemCounter[elemType] > 0:
                    hopout.info( ELEMTYPE(elemType) + ': {:12d}'.format(elemCounter[elemType]))
            hopout.sep()
            hopout.routine('Writing HDF5 mesh to "{}"'.format(fname))

            with h5py.File(fname, mode='w') as f:
                # Store same basic information
                common = Common()
                f.attrs['HoprVersion'   ] = common.version
                f.attrs['HoprVersionInt'] = common.__version__.micro + common.__version__.minor*100 + common.__version__.major*10000

                # Store mesh information
                f.attrs['Ngeo'          ] = mesh_vars.nGeo
                f.attrs['nElems'        ] = nElems
                f.attrs['nSides'        ] = nSides
                f.attrs['nNodes'        ] = nNodes
                f.attrs['nUniqueSides'  ] = np.max(sideInfo[:, 1])
                f.attrs['nUniqueNodes'  ] = np.max(nodeInfo)

                _ = f.create_dataset('ElemInfo'     , data=elemInfo)
                _ = f.create_dataset('ElemCounter'  , data=np.array(list(elemCounter.items()), dtype=np.int32))
                _ = f.create_dataset('SideInfo'     , data=sideInfo)
                _ = f.create_dataset('GlobalNodeIDs', data=nodeInfo)
                _ = f.create_dataset('NodeCoords'   , data=nodeCoords)

                if FEMElemInfo is not None:
                    f.attrs['FEMconnect'] = 'ON'
                    _ = f.create_dataset('FEMElemInfo'       , data=FEMElemInfo)
                    _ = f.create_dataset('VertexInfo'        , data=vertexInfo)
                    _ = f.create_dataset('VertexConnectInfo' , data=vertexConnectInfo)
                    _ = f.create_dataset('EdgeInfo'          , data=edgeInfo)
                    _ = f.create_dataset('EdgeConnectInfo'   , data=edgeConnectInto)

                # Store boundary information
                f.attrs['nBCs'          ] = nBCs
                bcNames = [f'{s.name:<255}' for s in bcs]
                bcTypes = np.zeros((nBCs, 4), dtype=np.int32)
                for iBC, bc in enumerate(bcs):
                    bcTypes[iBC, :] = bc.type

                _ = f.create_dataset('BCNames'   , data=np.array(bcNames, dtype='S'))
                _ = f.create_dataset('BCType'    , data=bcTypes)

                # Check if there is a periodic vector and write it to mesh file
                nVV = len(mesh_vars.vvs)
                if nVV > 0:
                    vvs = np.zeros((nVV, 3), dtype=np.float64)
                    for iVV, vv in enumerate(mesh_vars.vvs):
                        vvs[iVV, :] = vv['Dir']
                    _ = f.create_dataset('VV', data=vvs)

        case MeshFormat.FORMAT_VTK:
            mesh  = mesh_vars.mesh
            pname = io_vars.projectname
            fname = '{}_mesh.vtk'.format(pname)

            hopout.sep()
            hopout.routine('Writing VTK mesh to "{}"'.format(fname))

            mesh.write(fname, file_format='vtk42')

        case MeshFormat.FORMAT_GMSH:
            mesh  = mesh_vars.mesh
            pname = io_vars.projectname
            fname = '{}_mesh.msh'.format(pname)

            # Mixed elements required gmsh:dim_tags
            # > FIXME: THIS ARE DUMMY ENTRIES AND ONLY GENERATE A POINT MESH
            mesh.point_data.update({'gmsh:dim_tags': np.array([[0, i] for i in range(len(mesh.points))])})

            # Mixed elements require gmsh:physical and gmsh:geometrical
            # > FIXME: THIS ARE DUMMY ENTRIES AND ONLY GENERATE A POINT MESH
            cell_types = mesh.cells_dict.keys()
            cell_data  = [np.ones(mesh.cells_dict[cell_type].data.shape[1], dtype=int) for cell_type in cell_types
]
            mesh.cell_data_dict.update({'gmsh:physical':    cell_data})
            mesh.cell_data_dict.update({'gmsh:geometrical': cell_data})

            hopout.sep()
            hopout.routine('Writing GMSH mesh to "{}"'.format(fname))

            hopout.warning('GMSH output is not yet fully supported, only a point mesh is generated!')

            mesh.write(fname, file_format='gmsh')

        case _:  # Default
            hopout.warning('Unknown output format {}, exiting...'.format(io_vars.outputformat))
            sys.exit(1)

    # hopout.sep()
    # hopout.info('OUTPUT MESH DONE!')


def getMeshInfo() -> tuple[np.ndarray,         # ElemInfo
                           np.ndarray,         # SideInfo
                           np.ndarray,         # NodeInfo
                           np.ndarray,         # NodeCoords
                           np.ndarray | None,  # Optional[FEMElemInfo]
                           np.ndarray | None,  # Optional[VertexInfo]
                           np.ndarray | None,  # Optional[VertexConnectInfo]
                           np.ndarray | None,  # Optional[EdgeInfo]
                           np.ndarray | None,  # Optional[EdgeConnectInfo]
                           dict[int, int]
                          ]:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.mesh.fem.fem import getFEMInfo
    from pyhope.mesh.mesh_common import LINTEN
    from pyhope.io.io_vars import ELEM, SIDE
    # ------------------------------------------------------

    mesh:   Final             = mesh_vars.mesh
    elems:  Final[list]       = mesh_vars.elems
    sides:  Final[list]       = mesh_vars.sides
    points: Final[np.ndarray] = mesh.points

    nElems: Final[int] = len(elems)
    nSides: Final[int] = len(sides)
    nNodes: Final[int] = np.sum([s.nodes.size for s in elems])  # number of non-unique nodes

    # Create the ElemCounter
    elemCounter = OrderedDict()
    for elemType in ELEM.TYPES:
        elemCounter[elemType] = 0

    # Fill the ElemInfo
    elemInfo  = np.zeros((nElems, ELEM.INFOSIZE), dtype=np.int32)
    sideCount = 0  # elem['Sides'] might work as well
    nodeCount = 0  # elem['Nodes'] contains the unique nodes

    for iElem, elem in enumerate(elems):
        elemInfo[iElem, ELEM.TYPE     ] = elem.type
        elemInfo[iElem, ELEM.ZONE     ] = 1  # FIXME

        elemInfo[iElem, ELEM.FIRSTSIDE] = sideCount
        elemInfo[iElem, ELEM.LASTSIDE ] = sideCount + len(elem.sides)
        sideCount += len(elem.sides)

        elemInfo[iElem, ELEM.FIRSTNODE] = nodeCount
        elemInfo[iElem, ELEM.LASTNODE ] = nodeCount + len(elem.nodes)
        nodeCount += len(elem.nodes)

        elemCounter[elem.type] += 1

    # Set the global side ID
    globalSideID     = 0
    highestSideID    = 0
    usedSideIDs      = set()  # Set to track used side IDs
    availableSideIDs = []     # Min-heap for gap

    for iSide, side in enumerate(sides):
        # Already counted the side
        if side.globalSideID is not None:
            continue

        # Get the smallest available globalSideID from the heap, if any
        if availableSideIDs:
            globalSideID = heapq.heappop(availableSideIDs)
        else:
            # Use the current maximum ID and increment
            globalSideID = highestSideID + 1

        # Mark the side ID as used
        highestSideID = max(globalSideID, highestSideID)
        usedSideIDs.add(globalSideID)
        # side.update(globalSideID=globalSideID)
        side.globalSideID = globalSideID

        if side.connection is None:         # BC side
            pass
        elif side.connection < 0:           # Big mortar side
            pass
        elif side.MS == 1:                  # Internal / periodic side (master side)
            # Get the connected slave side
            nbSideID = side.connection

            # Reclaim the ID of the slave side if already assigned
            if sides[nbSideID].globalSideID is not None:
                reclaimedID = sides[nbSideID].globalSideID
                usedSideIDs.remove(reclaimedID)
                heapq.heappush(availableSideIDs, reclaimedID)

            # Set the negative globalSideID of the slave side
            # sides[nbSideID].update(globalSideID=-(globalSideID))
            sides[nbSideID].globalSideID = -(globalSideID)

    # If there are any gaps in the side IDs, fill them by reassigning consecutive values
    if availableSideIDs:
        # Collect all master sides (globalSideID > 0) and sort them by their current IDs
        masters = sorted((side for side in sides if side.globalSideID > 0), key=lambda side: side.globalSideID)

        # Build a mapping from old master ID to new consecutive IDs (starting at 1)
        mapping = {side.globalSideID: newID for newID, side in enumerate(masters, start=1)}

        # Update the sides based on the mapping
        for side in sides:
            # For slave sides, update to the negative of the mapped master ID
            side.globalSideID = mapping[side.globalSideID] if side.globalSideID > 0 else -mapping[-side.globalSideID]

    # Fill the SideInfo
    sideInfo  = np.zeros((nSides, SIDE.INFOSIZE), dtype=np.int32)

    for iSide, side in enumerate(sides):
        sideInfo[iSide, SIDE.TYPE     ] = side.sideType
        sideInfo[iSide, SIDE.ID       ] = side.globalSideID
        # Connected sides
        if side.connection is None:                                # BC side
            sideInfo[iSide, SIDE.NBELEMID      ] = 0
            sideInfo[iSide, SIDE.NBLOCSIDE_FLIP] = 0
            sideInfo[iSide, SIDE.BCID          ] = side.bcid + 1
        elif side.locMortar is not None:                           # Small mortar side
            nbSideID = side.connection
            nbElemID = sides[nbSideID].elemID + 1  # Python -> HOPR index
            sideInfo[iSide, SIDE.NBELEMID      ] = nbElemID
        elif side.connection is not None and side.connection < 0:  # Big mortar side
            # WARNING: This is not a sideID, but the mortar type
            sideInfo[iSide, SIDE.NBELEMID      ] = side.connection
            # Periodic mortar sisters have a BCID
            if side.bcid is not None:
                sideInfo[iSide, SIDE.BCID      ] = side.bcid + 1
        else:                                                      # Internal side
            nbSideID = side.connection
            nbElemID = sides[nbSideID].elemID + 1  # Python -> HOPR index
            sideInfo[iSide, SIDE.NBELEMID      ] = nbElemID
            if side.sideType < 0:  # Small mortar side
                sideInfo[iSide, SIDE.NBLOCSIDE_FLIP] = side.flip
            elif side.flip == 0:     # Master side
                sideInfo[iSide, SIDE.NBLOCSIDE_FLIP] = sides[nbSideID].locSide*10
            else:
                sideInfo[iSide, SIDE.NBLOCSIDE_FLIP] = sides[nbSideID].locSide*10 + side.flip

            # Periodic/inner sides still have a BCID
            if side.bcid is not None:
                sideInfo[iSide, SIDE.BCID      ] = side.bcid + 1
            else:
                sideInfo[iSide, SIDE.BCID      ] = 0

    # Fill the NodeInfo
    nodeInfo   = np.zeros((nNodes)   , dtype=np.int32)

    # Fill the NodeCoords
    nodeCoords = np.zeros((nNodes, 3), dtype=np.float64)
    nodeCount  = 0

    # Cache the mapping
    linCache   = {}

    for iElem, elem in enumerate(elems):
        # Mesh coordinates are stored in meshIO sorting
        elemType = elem.type
        if elemType in linCache:
            mapLin = linCache[elemType]
        else:
            _, mapLin = LINTEN(elem.type, order=mesh_vars.nGeo)
            mapLin    = np.array(tuple(mapLin[np.int64(i)] for i in range(len(mapLin))))
            linCache[elemType] = mapLin

        elemNodes  = np.asarray(elem.nodes)
        nElemNodes = elemNodes.size
        indices    = nodeCount + mapLin[:nElemNodes]

        # Assign nodeInfo and nodeCoords in vectorized fashion
        nodeInfo[  indices   ] = elemNodes + 1
        nodeCoords[indices, :] = points[elemNodes]

        nodeCount += nElemNodes

    if hasattr(elems[0], 'vertexInfo') and elems[0].vertexInfo is not None:
        FEMElemInfo, vertexInfo, vertexConnectInfo, edgeInfo, edgeConnectInto = getFEMInfo(nodeInfo)
    else:
        FEMElemInfo, vertexInfo, vertexConnectInfo, edgeInfo, edgeConnectInto = [None for _ in range(5)]

    return elemInfo, sideInfo, nodeInfo, nodeCoords, \
           FEMElemInfo, vertexInfo, vertexConnectInfo, edgeInfo, edgeConnectInto, \
           elemCounter
